use hmac::{Hmac, Mac};
use jwt::{SignWithKey, VerifyWithKey};
use poem::{
    error::InternalServerError, listener::TcpListener, web::Data, EndpointExt, Request, Result,
    Route,
};
use poem_openapi::{
    auth::ApiKey,
    param::Query,
    payload::{Json, PlainText},
    Object, OpenApi, OpenApiService, SecurityScheme,
};
use serde::{Deserialize, Serialize};
use sha2::Sha256;

const SERVER_KEY: &[u8] = b"13371337";
type ServerKey = Hmac<Sha256>;

/// API Auth
#[derive(SecurityScheme)]
#[oai(
    type = "api_key",
    key_name = "X-API-Key",
    in = "header",
    checker = "api_checker"
)]
struct ApiKeyAuth(User);

async fn api_checker(req: &Request, api_key: ApiKey) -> Option<User> {
    let server_key = req.data::<ServerKey>().unwrap();
    VerifyWithKey::<User>::verify_with_key(api_key.key.as_str(), server_key).ok()
}

#[derive(Debug, Serialize, Deserialize)]
struct User {
    login: String,
    name: String,
}

#[derive(Object)]
struct LoginRequest {
    login: String,
    name: String,
}

struct Api;

#[OpenApi]
impl Api {
    #[oai(path = "/hello", method = "get")]
    async fn hello(&self, name: Query<Option<String>>) -> PlainText<String> {
        match name.0 {
            Some(name) => PlainText(format!("Hello {}!", name)),
            None => PlainText(String::from("Hello!")),
        }
    }

    #[oai(path = "/user", method = "get")]
    async fn user(&self, auth: ApiKeyAuth) -> PlainText<String> {
        PlainText(format!(
            "Hello {}. Your real name is {}!",
            auth.0.login, auth.0.name
        ))
    }

    #[oai(path = "/login", method = "post")]
    async fn login(
        &self,
        server_key: Data<&ServerKey>,
        req: Json<LoginRequest>,
    ) -> Result<PlainText<String>> {
        let token = User {
            login: req.0.login,
            name: req.0.name,
        }
        .sign_with_key(server_key.0)
        .map_err(InternalServerError)?;
        Ok(PlainText(token))
    }
}

#[tokio::main]
async fn main() -> Result<(), std::io::Error> {
    let api_service =
        OpenApiService::new(Api, "Test Api", "1.3.37").server("http://localhost:6969/api");
    let ui = api_service.swagger_ui();
    let spec = api_service.spec_endpoint();
    let server_key = ServerKey::new_from_slice(SERVER_KEY).unwrap();
    let app = Route::new()
        .nest("/api", api_service)
        .nest("/swagger", ui)
        .nest("/openapi", spec)
        .data(server_key);

    poem::Server::new(TcpListener::bind("127.0.0.1:6969"))
        .run(app)
        .await
}
