use std::{env, sync::Arc};

use cassandra_cpp::{stmt, Session};

use crate::repo::Result;
use crate::model::todo::Todo;

pub struct TodoRepository {
    session: Arc<Session>,
    table: String,
}

impl TodoRepository {
    pub fn new(session: Arc<Session>) -> TodoRepository {
        let keyspace_name =
            env::var("CASSANDRA_KEYSPACE_NAME").expect("Value should be definitely set in init");
        let table = format!("{keyspace_name}.todo");
        let query = stmt!(&format!(
            "CREATE TABLE IF NOT EXISTS {table} (
                id uuid,
                user_id uuid,
                title text,
                priority varchar,
                status varchar,
                description text,
                PRIMARY KEY (id, user_id)
            );"
        ));

        session
            .execute(&query)
            .wait()
            .expect("Should create todo table if not exists");

        // Create a secondary index so we can also query for todos of specific users
        // without the use of ALLOW FILTERING (because this might be slow)
        let query = stmt!(&format!("CREATE INDEX IF NOT EXISTS ON {table} (user_id);"));

        session
            .execute(&query)
            .wait()
            .expect("Should create secondary index for user_id column");

        TodoRepository { session, table }
    }

    /// Creates OR updates the provided todo.
    pub fn create(&self, todo: &Todo) -> Result<()> {
        // We can just create todos, nothing to check beforehand
        // Trying out the JSON syntax
        // This is quite nice since cassandra just updates the record if
        // the primary key (id, user_id) are already present
        let json = todo.to_json()?;
        log::debug!("Got this json for the query: {json:?}");
        let query = stmt!(&format!("INSERT INTO {} JSON '{}';", self.table, json));

        self.session.execute(&query).wait()?;
        Ok(())
    }

    // REMARK: Not required right now
    // pub fn read(&self, id: &uuid::Uuid) -> Result<Option<Todo>> {
    //     let uuid = cassandra_cpp::Uuid::from(id.clone());

    //     let mut query = stmt!(&format!("SELECT JSON * FROM {} WHERE id = ?", self.table));
    //     query.bind_uuid(0, uuid).expect("Binds id");

    //     Ok(match self.session.execute(&query).wait()?.first_row() {
    //         Some(row) => Some(Todo::from_json(&format!("{row}"))),
    //         None => None,
    //     })
    // }

    // REMARK: Not required right now
    // pub fn read_all(&self) -> Result<Vec<Todo>> {
    //     let query = stmt!(&format!("SELECT JSON * FROM {}", self.table));
    //     Ok(self
    //         .session
    //         .execute(&query)
    //         .wait()?
    //         .iter()
    //         .map(|json| Todo::from_json(&format!("{json}")))
    //         .collect())
    // }

    pub fn read_all_by_user_id(&self, user_id: &uuid::Uuid) -> Result<Vec<Todo>> {
        let mut query = stmt!(&format!(
            "SELECT JSON * FROM {} WHERE user_id = ?",
            self.table
        ));

        let uuid = cassandra_cpp::Uuid::from(*user_id);
        query.bind_uuid(0, uuid).expect("Bind user_id");

        Ok(self
            .session
            .execute(&query)
            .wait()?
            .iter()
            .map(|json| Todo::from_json(&format!("{json}")))
            .collect())
    }

    // REMARK: Not required right now
    // pub fn update(&self, todo: &Todo) -> Result<()> {
    //     // Read the comment in create
    //     self.create(todo)
    // }

    pub fn delete(&self, id: &uuid::Uuid, user_id: &uuid::Uuid) -> Result<()> {

        let mut query = stmt!(&format!(
            "DELETE FROM {} WHERE id = ? AND user_id = ?",
            self.table
        ));

        let id_uuid = cassandra_cpp::Uuid::from(*id);
        let user_id_uuid = cassandra_cpp::Uuid::from(*user_id);
        query.bind_uuid(0, id_uuid).expect("Bind id");
        query.bind_uuid(1, user_id_uuid).expect("Bind user_id");

        self.session.execute(&query).wait()?;
        Ok(())
    }
}
