use std::{env, error::Error, fmt::Display, sync::Arc};

use cassandra_cpp::{stmt, Session};

use crate::model::user::User;
use crate::repo::Result;

// ---------------------------------------------------------------------------------
//                                  Errors
// ---------------------------------------------------------------------------------

#[derive(Debug)]
struct UserAlreadyExistsError;
#[derive(Debug)]
struct UserDoesNotExistError;

impl Display for UserAlreadyExistsError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "User already exists!")
    }
}

impl Display for UserDoesNotExistError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "User does not exists!")
    }
}

impl Error for UserAlreadyExistsError {}
impl Error for UserDoesNotExistError {}

// ---------------------------------------------------------------------------------
//                                  Code
// ---------------------------------------------------------------------------------

pub struct UserRepository {
    session: Arc<Session>,
    table: String,
}

/// This is an abstraction for the cassandra db used here.
/// It provides the typical CRUD actions.
impl UserRepository {
    pub fn new(session: Arc<Session>) -> UserRepository {
        let keyspace_name =
            env::var("CASSANDRA_KEYSPACE_NAME").expect("Value should be definitely set in init");
        let table = format!("{keyspace_name}.user");
        let query = stmt!(&format!(
            "CREATE TABLE IF NOT EXISTS {table} (
                id uuid,
                login text,
                hash ascii,
                salt text,
                PRIMARY KEY (id)
            );"
        ));

        session
            .execute(&query)
            .wait()
            .expect("Should create user table if not exists");

        let query = stmt!(&format!("CREATE INDEX IF NOT EXISTS ON {table} (login);"));

        session
            .execute(&query)
            .wait()
            .expect("Should create secondary index for login column");

        UserRepository { session, table }
    }

    pub fn create(&self, user: &User) -> Result<()> {
        if self.read_by_login(user.login())?.is_some() {
            return Err(Box::new(UserAlreadyExistsError));
        }

        let mut query = stmt!(&format!(
            "INSERT INTO {} (id, login, hash, salt)
             VALUES (?, ?, ?, ?);",
            self.table
        ));

        let uuid = cassandra_cpp::Uuid::from(*user.id());
        query.bind_uuid(0, uuid).expect("Binds the id");
        query.bind_string(1, user.login()).expect("Binds the login");
        query.bind_string(2, user.hash()).expect("Binds the hash");
        query.bind_string(3, user.salt()).expect("Binds the salt");
        self.session.execute(&query).wait()?;
        Ok(())
    }

    pub fn read(&self, id: &uuid::Uuid) -> Result<Option<User>> {
        let uuid = cassandra_cpp::Uuid::from(*id);
        // There is no OR in cassandra statements
        let mut query = stmt!(&format!("SELECT JSON * FROM {} WHERE id = ?", self.table));
        query.bind_uuid(0, uuid).expect("Binds id");

        Ok(self
            .session
            .execute(&query)
            .wait()?
            .first_row()
            .map(|row| User::from_json(&format!("{row}"))))
    }

    pub fn read_by_login(&self, login: &str) -> Result<Option<User>> {
        let mut query = stmt!(&format!(
            "SELECT JSON * FROM {} WHERE login = ?",
            self.table
        ));
        query.bind_string(0, login).expect("Binds login");
        Ok(self
            .session
            .execute(&query)
            .wait()?
            .first_row()
            .map(|row| User::from_json(&format!("{row}"))))
    }

    pub fn read_all(&self) -> Result<Vec<User>> {
        let query = stmt!(&format!("SELECT JSON * FROM {};", self.table));
        Ok(self
            .session
            .execute(&query)
            .wait()?
            .iter()
            .map(|json| User::from_json(&format!("{json}")))
            .collect())
    }

    pub fn update(&self, user: &User) -> Result<()> {
        match self.read(user.id())? {
            Some(u) => {
                log::info!("Modifying {u:?} to represent {user:?}");

                let mut query = stmt!(&format!(
                    "UPDATE {} SET login = ?, hash = ?, salt = ?
                     WHERE id = ?;",
                    self.table
                ));

                let uuid = cassandra_cpp::Uuid::from(*user.id());
                query.bind_string(0, user.login()).expect("Binds the login");
                query.bind_string(1, user.hash()).expect("Binds the hash");
                query.bind_string(2, user.salt()).expect("Binds the salt");
                query.bind_uuid(3, uuid).expect("Binds the id");
                self.session.execute(&query).wait()?;
                Ok(())
            }
            None => Err(Box::new(UserDoesNotExistError)),
        }
    }
}
