use actix_identity::Identity;
use actix_session::Session;
use actix_web::{
    delete, get, put,
    web::{self, Json},
    HttpResponse, Responder,
};

use crate::{
    model::todo::Todo, repo::todo_repository::TodoRepository, routes::get_user_from_session,
};

// IDEA: Depending on role see all todos not just your own (Group todos)
/// Fetches a list of all todos belonging to the currently logged in user.
#[get("/todo")]
pub async fn get_todo(
    _id: Identity,
    session: Session,
    repo: web::Data<TodoRepository>,
) -> impl Responder {
    match get_user_from_session(session) {
        Some(user) => match repo.read_all_by_user_id(user.id()) {
            Ok(todos) => HttpResponse::Ok().json(todos),
            Err(err) => {
                log::warn!("Could not fetch todos: {err}");
                HttpResponse::NoContent().finish()
            }
        },
        None => HttpResponse::BadRequest().finish(),
    }
}

// IDEA: Let users create todos for other users if role allows it
/// Creates or updates the todo provided in the body of the request.
#[put("/todo")]
pub async fn put_todo(
    _id: Identity,
    session: Session,
    repo: web::Data<TodoRepository>,
    todo: Json<Todo>,
) -> impl Responder {
    let mut todo = todo; // To set user_id from session
    match get_user_from_session(session) {
        Some(user) => {
            todo.set_user_id(*user.id());
            match repo.create(&todo) {
                Ok(_) => HttpResponse::Ok().finish(),
                Err(err) => {
                    log::warn!("Error while creating or updating {todo:?}. ERROR: {err}");
                    HttpResponse::BadRequest().finish()
                }
            }
        }
        None => HttpResponse::BadRequest().finish(),
    }
}

/// Deletes a todo via its id which is provided in the url.
/// Example `http://localhost/todo/uuid-of-the-todo-to-delete`.
#[delete("/todo/{todo_id}")]
pub async fn delete_todo(
    _id: Identity,
    session: Session,
    repo: web::Data<TodoRepository>,
    todo_id: web::Path<uuid::Uuid>,
) -> impl Responder {
    match get_user_from_session(session) {
        Some(user) => {
            match repo.delete(&todo_id, user.id()) {
                Ok(_) => HttpResponse::Ok().finish(),
                Err(err) => {
                    log::warn!("Error while deleting todo with id {todo_id}: {err}");
                    HttpResponse::BadRequest().finish()
                }
            }
        }
        None => HttpResponse::BadRequest().finish(),
    }
}
