import { Component, createSignal, onMount, Show } from 'solid-js';
import { v4 as uuidv4 } from 'uuid';
import RestClient from '../api/RestClient';
import { user } from '../App';
import Button from '../ui/Button';
import Table from '../ui/Table';

export type Todo = {
  id: string;
  user_id: string;
  title: string;
  priority: Priority;
  status: Status;
  description: string;
};

type TodoModalProps = {
  todo: Todo;
};

export enum Status {
  Todo = 'Todo',
  Doing = 'Doing',
  Done = 'Done',
}

export enum Priority {
  High = 'High',
  Normal = 'Normal',
  Low = 'Low',
}

const createNewTodo = (): Todo => ({
  id: uuidv4(),
  user_id: user().id,
  title: 'New Title',
  description: 'Some description.',
  status: Status.Todo,
  priority: Priority.Normal,
});

const TodoModal: Component<TodoModalProps> = (props) => {
  const [show, setShow] = createSignal(false);
  const [todo, setTodo] = createSignal(props.todo);
  const inputClass =
    'form-control block w-full px-3 py-1.5 text-base font-normal text-gray-700 bg-white bg-clip-padding border border-solid border-gray-300 rounded transition ease-in-out m-0 focus:text-gray-700 focus:bg-white focus:border-blue-600 focus:outline-none';
  const selectClass =
    'form-select block w-full px-3 py-1.5 text-base font-normal text-gray-700 bg-white bg-clip-padding bg-no-repeat border border-solid border-gray-300 rounded transition ease-in-out m-0 focus:text-gray-700 focus:bg-white focus:border-blue-600 focus:outline-none';
  const labelClass = 'form-label inline-block mb-2 text-gray-700';
  console.log(Object.is(props.todo, todo()));
  return (
    <>
      <Show
        when={Object.is(props.todo, todo())}
        fallback={
          <>
            <Button backgroundColor='magenta' onClick={() => setShow(true)}>
              🖉
            </Button>
          </>
        }
      >
        <Button onClick={() => setShow(true)}>🖉</Button>
      </Show>
      <Show when={show()} fallback={<></>}>
        <div class='fixed top-0 left-0 w-screen h-screen flex items-center justify-center bg-sh-bgP2 bg-opacity-50 transform transition-transform duration-300'>
          <div class='p-6 rounded-lg bg-white w-1/2'>
            <form>
              <div class='form-group mb-6'>
                <label for='title' class={labelClass}>
                  Title
                </label>
                <input
                  type='text'
                  class={inputClass}
                  id='title'
                  onInput={(e) =>
                    setTodo((t) => ({ ...t, title: e.currentTarget.value } as Todo))
                  }
                  value={todo()?.title}
                />
              </div>
              <div class='form-group mb-6'>
                <label for='description' class={labelClass}>
                  Description
                </label>
                <textarea
                  class={inputClass}
                  id='description'
                  onInput={(e) =>
                    setTodo((t) => ({ ...t, description: e.currentTarget.value } as Todo))
                  }
                  value={todo()?.description}
                />
              </div>
              <div class='form-group mb-6'>
                <label for='status' class={labelClass}>
                  Status
                </label>
                <select
                  onChange={(e) =>
                    setTodo((t) => ({ ...t, status: e.currentTarget.value } as Todo))
                  }
                  class={selectClass}
                >
                  {Object.keys(Status).map((k) => {
                    if (todo().status === Status[k as Status]) {
                      return (
                        <option selected value={Status[k as Status]}>
                          {k}
                        </option>
                      );
                    } else {
                      return <option value={Status[k as Status]}>{k}</option>;
                    }
                  })}
                </select>
                <div class='form-group mb-6'>
                  <label for='priority' class={labelClass}>
                    Priority
                  </label>
                  <select
                    onChange={(e) =>
                      setTodo((t) => ({ ...t, priority: e.currentTarget.value } as Todo))
                    }
                    class={selectClass}
                  >
                    {Object.keys(Priority).map((k) => {
                      if (todo().priority === Priority[k as Priority]) {
                        return (
                          <option selected value={Priority[k as Priority]}>
                            {k}
                          </option>
                        );
                      } else {
                        return <option value={Priority[k as Priority]}>{k}</option>;
                      }
                    })}
                  </select>
                </div>
              </div>
            </form>
            <Button
              backgroundColor='green'
              onClick={() =>
                RestClient.PUT('/todo', JSON.stringify(todo())).then(() => {
                  fetchTodos();
                  setShow(false);
                })
              }
            >
              Save
            </Button>{' '}
            <Button backgroundColor='gray' onClick={() => setShow(false)}>
              Close
            </Button>{' '}
            <Button
              backgroundColor='red'
              onClick={() => {
                setTodo({ ...props.todo } as Todo);
                setShow(false);
              }}
            >
              Cancel
            </Button>
          </div>
        </div>
      </Show>
    </>
  );
};

const fetchTodos = async () => {
  setTodos((await RestClient.GET('/todo')) as Todo[]);
};

const [todos, setTodos] = createSignal<Todo[]>([]);

const Home: Component = () => {
  onMount(fetchTodos);
  return (
    <>
      <Show when={todos().length > 0} fallback={<></>}>
        <Table
          data={todos().map((todo) => ({
            Title: todo.title,
            Description: todo.description,
            Status: todo.status,
            Priority: todo.priority,
            Remove: (
              <Button
                backgroundColor='red'
                onClick={() => {
                  RestClient.DELETE(`/todo/${todo.id}`).then(() => fetchTodos());
                }}
              >
                🗑
              </Button>
            ),
            Edit: <TodoModal todo={todo}></TodoModal>,
          }))}
        ></Table>
      </Show>
      <div class='mt-10 ml-10'>
        <Button
          backgroundColor='green'
          onClick={() =>
            RestClient.PUT('/todo', JSON.stringify(createNewTodo())).then(() => {
              fetchTodos();
            })
          }
        >
          New
        </Button>
      </div>
    </>
  );
};

export default Home;
